@@ -1,6 +1,6 @@
 use anyhow::{anyhow, Result};
 use cairo::{Antialias, Context, Format, ImageSurface, Surface};
-use chrono::{Local, Locale, Timelike};
+use chrono::{Local, Locale, Timelike, format::{StrftimeItems, Item as ChronoItem}};
 use drm::control::ClipRect;
 use freedesktop_icons::lookup;
 use input::{
@@ -27,7 +27,7 @@ use privdrop::PrivDrop;
 use std::{
     cmp::min,
     collections::HashMap,
-    fs::{File, OpenOptions},
+    fs::{self, File, OpenOptions},
     os::{
         fd::{AsFd, AsRawFd},
         unix::{fs::OpenOptionsExt, io::OwnedFd},
@@ -35,6 +35,7 @@ use std::{
     panic::{self, AssertUnwindSafe},
     path::{Path, PathBuf},
 };
+use udev::MonitorBuilder;
 
 mod backlight;
 mod config;
@@ -54,11 +55,41 @@ const BUTTON_COLOR_ACTIVE: f64 = 0.400;
 const ICON_SIZE: i32 = 48;
 const TIMEOUT_MS: i32 = 10 * 1000;
 
+#[derive(Clone, Copy, PartialEq, Eq)]
+enum BatteryState {
+    NotCharging,
+    Charging,
+    Low,
+}
+
+struct BatteryImages {
+    plain: Vec<Handle>,
+    charging: Vec<Handle>,
+    bolt: Handle,
+}
+
+#[derive(Eq, PartialEq, Copy, Clone)]
+enum BatteryIconMode {
+    Percentage,
+    Icon,
+    Both
+}
+
+impl BatteryIconMode {
+    fn should_draw_icon(self) -> bool {
+        self != BatteryIconMode::Percentage
+    }
+    fn should_draw_text(self) -> bool {
+        self != BatteryIconMode::Icon
+    }
+}
+
 enum ButtonImage {
     Text(String),
     Svg(Handle),
     Bitmap(ImageSurface),
-    Time(String, String),
+    Time(Vec<ChronoItem<'static>>, Locale),
+    Battery(String, BatteryIconMode, BatteryImages),
 }
 
 struct Button {
@@ -152,6 +183,62 @@ fn try_load_image(name: impl AsRef<str>, theme: Option<impl AsRef<str>>) -> Resu
     Err(last_err.context(format!("failed loading all possible paths for icon {name}")))
 }
 
+fn find_battery_device() -> Option<String> {
+    let power_supply_path = "/sys/class/power_supply";
+    if let Ok(entries) = fs::read_dir(power_supply_path) {
+        for entry in entries.flatten() {
+            let dev_path = entry.path();
+            let type_path = dev_path.join("type");
+            if let Ok(typ) = fs::read_to_string(&type_path) {
+                if typ.trim() == "Battery" {
+                    if let Some(name) = dev_path.file_name().and_then(|n| n.to_str()) {
+                        return Some(name.to_string());
+                    }
+                }
+            }
+        }
+    }
+    None
+}
+
+fn get_battery_state(battery: &str) -> (u32, BatteryState) {
+    let status_path = format!("/sys/class/power_supply/{}/status", battery);
+    let status = fs::read_to_string(&status_path)
+        .unwrap_or_else(|_| "Unknown".to_string());
+
+    #[cfg(target_arch = "x86_64")]
+    let capacity = {
+        let charge_now_path = format!("/sys/class/power_supply/{}/charge_now", battery);
+        let charge_full_path = format!("/sys/class/power_supply/{}/charge_full", battery);
+        let charge_now = fs::read_to_string(&charge_now_path)
+            .ok()
+            .and_then(|s| s.trim().parse::<f64>().ok());
+        let charge_full = fs::read_to_string(&charge_full_path)
+            .ok()
+            .and_then(|s| s.trim().parse::<f64>().ok());
+        match (charge_now, charge_full) {
+            (Some(now), Some(full)) if full > 0.0 => ((now / full) * 100.0).round() as u32,
+            _ => 100,
+        }
+    };
+
+    #[cfg(target_arch = "aarch64")]
+    let capacity = {
+        let capacity_path = format!("/sys/class/power_supply/{}/capacity", battery);
+        fs::read_to_string(&capacity_path)
+            .ok()
+            .and_then(|s| s.trim().parse::<u32>().ok())
+            .unwrap_or(100)
+    };
+
+    let status = match status.trim() {
+        "Charging" | "Full" => BatteryState::Charging,
+        "Discharging" if capacity < 10 => BatteryState::Low,
+        _ => BatteryState::NotCharging,
+    };
+    (capacity, status)
+}
+
 impl Button {
     fn with_config(cfg: ButtonConfig) -> Button {
         if let Some(text) = cfg.text {
@@ -159,11 +246,13 @@ impl Button {
         } else if let Some(icon) = cfg.icon {
             Button::new_icon(&icon, cfg.theme, cfg.action)
         } else if let Some(time) = cfg.time {
-            let locale = match cfg.locale {
-                 Some(l) => l,
-                 None => "POSIX".to_string()
-            };
-            Button::new_time(cfg.action, time, locale)
+            Button::new_time(cfg.action, &time, cfg.locale.as_deref())
+        } else if let Some(battery_mode) = cfg.battery {
+            if let Some(battery) = find_battery_device() {
+                Button::new_battery(cfg.action, battery, battery_mode, cfg.theme)
+            } else {
+                Button::new_text("Battery N/A".to_string(), cfg.action)
+            }
         } else {
             panic!("Invalid config, a button must have either Text, Icon or Time")
         }
@@ -185,13 +274,65 @@ impl Button {
             changed: false,
         }
     }
-
-    fn new_time(action: Key, format: String, locale: String) -> Button {
+    fn load_battery_image(icon: &str, theme: Option<impl AsRef<str>>) -> Handle {
+        if let ButtonImage::Svg(svg) = try_load_image(icon, theme).unwrap() {
+            return svg;
+        }
+        panic!("failed to load icon");
+    }
+    fn new_battery(action: Key, battery: String, battery_mode: String, theme: Option<impl AsRef<str>>) -> Button {
+        let bolt = Self::load_battery_image("bolt", theme.as_ref());
+        let mut plain = Vec::new();
+        let mut charging = Vec::new();
+        for icon in [
+            "battery_0_bar", "battery_1_bar", "battery_2_bar", "battery_3_bar",
+            "battery_4_bar", "battery_5_bar", "battery_6_bar", "battery_full",
+        ] {
+            plain.push(Self::load_battery_image(icon, theme.as_ref()));
+        }
+        for icon in [
+            "battery_charging_20", "battery_charging_30", "battery_charging_50",
+            "battery_charging_60", "battery_charging_80",
+            "battery_charging_90", "battery_charging_full",
+        ] {
+            charging.push(Self::load_battery_image(icon, theme.as_ref()));
+        }
+        let battery_mode = match battery_mode.as_str() {
+            "icon" => BatteryIconMode::Icon,
+            "percentage" => BatteryIconMode::Percentage,
+            "both" => BatteryIconMode::Both,
+            _ => panic!("invalid battery mode, accepted modes: icon, percentage, both"),
+        };
         Button {
             action,
             active: false,
             changed: false,
-            image: ButtonImage::Time(format, locale),
+            image: ButtonImage::Battery(battery, battery_mode, BatteryImages {
+                plain, bolt, charging
+            }),
+        }
+    }
+
+    fn new_time(action: Key, format: &str, locale_str: Option<&str>) -> Button {
+        let format_str = if format == "24hr" {
+            "%H:%M    %a %-e %b"
+        } else if format == "12hr" {
+            "%-l:%M %p    %a %-e %b"
+        } else {
+            format
+        };
+
+        let format_items = match StrftimeItems::new(format_str).parse_to_owned() {
+            Ok(s) => s,
+            Err(e) => panic!("Invalid time format, consult the configuration file for examples of correct ones: {e:?}"),
+        };
+
+        let locale = locale_str.and_then(|l| Locale::try_from(l).ok()).unwrap_or(Locale::POSIX);
+        Button {
+            action,
+            active: false,
+            changed: false,
+            image: ButtonImage::Time(format_items, locale),
         }
     }
     fn render(
@@ -228,36 +369,70 @@ impl Button {
                 c.fill().unwrap();
             }
             ButtonImage::Time(format, locale) => {
-                 let current_time = Local::now();
-                 let current_locale = Locale::try_from(locale.as_str()).unwrap_or(Locale::POSIX);
-                 let formatted_time;
-                 if format == "24hr" {
-                     formatted_time = format!(
-                     "{}:{}    {} {} {}",
-                      current_time.format_localized("%H", current_locale),
-                      current_time.format_localized("%M", current_locale),
-                      current_time.format_localized("%a", current_locale),
-                      current_time.format_localized("%-e", current_locale),
-                      current_time.format_localized("%b", current_locale)
-                 );
-                 } else {
-                     formatted_time = format!(
-                     "{}:{} {}    {} {} {}",
-                     current_time.format_localized("%-l", current_locale),
-                     current_time.format_localized("%M", current_locale),
-                     current_time.format_localized("%p", current_locale),
-                     current_time.format_localized("%a", current_locale),
-                     current_time.format_localized("%-e", current_locale),
-                     current_time.format_localized("%b", current_locale)
-                 );
-                 }
-                 let time_extents = c.text_extents(&formatted_time).unwrap();
-                 c.move_to(
-                     button_left_edge + (button_width as f64 / 2.0 - time_extents.width() / 2.0).round(),
-                     y_shift + (height as f64 / 2.0 + time_extents.height() / 2.0).round()
-                 );
-                 c.show_text(&formatted_time).unwrap();
-             }
+                let current_time = Local::now();
+                let formatted_time = current_time.format_localized_with_items(format.iter(), *locale).to_string();
+                let time_extents = c.text_extents(&formatted_time).unwrap();
+                c.move_to(
+                    button_left_edge + (button_width as f64 / 2.0 - time_extents.width() / 2.0).round(),
+                    y_shift + (height as f64 / 2.0 + time_extents.height() / 2.0).round()
+                );
+                c.show_text(&formatted_time).unwrap();
+            }
+            ButtonImage::Battery(battery, battery_mode, icons) => {
+                let (capacity, state) = get_battery_state(battery);
+                let icon = if battery_mode.should_draw_icon() {
+                    Some(match state {
+                        BatteryState::Charging => match capacity {
+                            0..=20 => &icons.charging[0],
+                            21..=30 => &icons.charging[1],
+                            31..=50 => &icons.charging[2],
+                            51..=60 => &icons.charging[3],
+                            61..=80 => &icons.charging[4],
+                            81..=99 => &icons.charging[5],
+                            _ => &icons.charging[6],
+                        },
+                        _ => match capacity {
+                            0 => &icons.plain[0],
+                            1..=20 => &icons.plain[1],
+                            21..=30 => &icons.plain[2],
+                            31..=50 => &icons.plain[3],
+                            51..=60 => &icons.plain[4],
+                            61..=80 => &icons.plain[5],
+                            81..=99 => &icons.plain[6],
+                            _ => &icons.plain[7],
+                        },
+                    })
+                } else if state == BatteryState::Charging {
+                    Some(&icons.bolt)
+                } else {
+                    None
+                };
+                let percent_str = format!("{:.0}%", capacity);
+                let extents = c.text_extents(&percent_str).unwrap();
+                let mut width = extents.width();
+                let mut text_offset = 0;
+                if let Some(svg) = icon {
+                    if !battery_mode.should_draw_text() {
+                        width = ICON_SIZE as f64;
+                    } else {
+                        width += ICON_SIZE as f64;
+                    }
+                    text_offset = ICON_SIZE;
+                    let x =
+                        button_left_edge + (button_width as f64 / 2.0 - width / 2.0).round();
+                    let y = y_shift + ((height as f64 - ICON_SIZE as f64) / 2.0).round();
+
+                    svg.render_document(c, &Rectangle::new(x, y, ICON_SIZE as f64, ICON_SIZE as f64))
+                        .unwrap();
+                }
+                if battery_mode.should_draw_text() {
+                    c.move_to(
+                        button_left_edge + (button_width as f64 / 2.0 - width / 2.0 + text_offset as f64).round(),
+                        y_shift + (height as f64 / 2.0 + extents.height() / 2.0).round(),
+                    );
+                    c.show_text(&percent_str).unwrap();
+                }
+            }
         }
     }
     fn set_active<F>(&mut self, uinput: &mut UInputHandle<F>, active: bool)
@@ -271,6 +446,18 @@ impl Button {
             toggle_key(uinput, self.action, active as i32);
         }
     }
+    fn set_backround_color(&self, c: &Context, color: f64) {
+        if let ButtonImage::Battery(battery, _, _) = &self.image {
+            let (_, state) = get_battery_state(battery);
+            match state {
+                BatteryState::NotCharging => c.set_source_rgb(color, color, color),
+                BatteryState::Charging => c.set_source_rgb(0.0, color, 0.0),
+                BatteryState::Low => c.set_source_rgb(color, 0.0, 0.0),
+            }
+        } else {
+            c.set_source_rgb(color, color, color);
+        }
+    }
 }
 
 #[derive(Default)]
@@ -296,9 +483,6 @@ impl FunctionLayer {
                         println!("Stretch value must be at least 1, setting to 1.");
                         stretch = 1;
                     }
-                    if cfg.time.is_some() {
-                        stretch = stretch * 3;
-                    }
                     **state += stretch;
                     Some((i, Button::with_config(cfg)))
                 })
@@ -383,7 +567,7 @@ impl FunctionLayer {
                 );
                 c.fill().unwrap();
             }
-            c.set_source_rgb(color, color, color);
+            button.set_backround_color(&c, color);
             // draw box with rounded corners
             c.new_sub_path();
             let left = left_edge + radius;
@@ -588,6 +772,12 @@ fn real_main(drm: &mut DrmBackend) {
     let mut input_main = Libinput::new_with_udev(Interface);
     input_tb.udev_assign_seat("seat-touchbar").unwrap();
     input_main.udev_assign_seat("seat0").unwrap();
+    let udev_monitor = MonitorBuilder::new()
+        .unwrap()
+        .match_subsystem("power_supply")
+        .unwrap()
+        .listen()
+        .unwrap();
     let epoll = Epoll::new(EpollCreateFlags::empty()).unwrap();
     epoll
         .add(input_main.as_fd(), EpollEvent::new(EpollFlags::EPOLLIN, 0))
@@ -598,6 +788,9 @@ fn real_main(drm: &mut DrmBackend) {
     epoll
         .add(cfg_mgr.fd(), EpollEvent::new(EpollFlags::EPOLLIN, 2))
         .unwrap();
+    epoll
+        .add(&udev_monitor, EpollEvent::new(EpollFlags::EPOLLIN, 3))
+        .unwrap();
     uinput.set_evbit(EventKind::Key).unwrap();
     for layer in &layers {
         for button in &layer.buttons {
@@ -649,6 +842,9 @@ fn real_main(drm: &mut DrmBackend) {
                  needs_complete_redraw = true;
                  last_redraw_minute = current_minute;
             }
+            if let ButtonImage::Battery(_, _, _) = button.1.image {
+                button.1.changed = true;
+            }
         }
 
         if needs_complete_redraw || layers[active_layer].buttons.iter().any(|b| b.1.changed) {
@@ -678,6 +874,9 @@ fn real_main(drm: &mut DrmBackend) {
             Err(Errno::EINTR) | Ok(_) => 0,
             e => e.unwrap(),
         };
+
+        _ = udev_monitor.iter().last();
+
         input_tb.dispatch().unwrap();
         input_main.dispatch().unwrap();
         for event in &mut input_tb.clone().chain(input_main.clone()) {
